unit pkcs11_session;

interface

uses
  pkcs11t,
  pkcs11_api,
  pkcs11_object,
  pkcs11_attribute,
  pkcs11_mechanism;

type
  TPKCS11UserType = (utUser, utSecurityOfficer, utContextSpecific);
const
  PKCS11_USER_TYPE : array [TPKCS11UserType] of CK_USER_TYPE = (
                              CKU_USER,
                              CKU_SO,
                              CKU_CONTEXT_SPECIFIC
                             );

type
  TPKCS11SessionState = (
                         ssROPublic,
                         ssROUserFunctions,
                         ssRWPublic,
                         ssRWUserFunctions,
                         ssRWSOFunctions
                        );
const
  PKCS11_SESSION_STATE : array [TPKCS11SessionState] of CK_STATE = (
                                  CKS_RO_PUBLIC_SESSION,
                                  CKS_RO_USER_FUNCTIONS,
                                  CKS_RW_PUBLIC_SESSION,
                                  CKS_RW_USER_FUNCTIONS,
                                  CKS_RW_SO_FUNCTIONS
                                 );

  PKCS11_SESSION_STATE_DESC : array [TPKCS11SessionState] of string = (
                                  'Read-only Public',
                                  'Read-only User',
                                  'Read/write Public',
                                  'Read/write User',
                                  'Read/write Security Officer'
                                 );

type
  TPKCS11Session = class(TPKCS11API)
  private
  protected
    FSlotID: integer;
    FLastEncryptDecryptLen: integer;

    // Returns TRUE/FALSE, depending on whether the session is open/valid or
    // not
    function GetOpen(): boolean;

    function GetState(): TPKCS11SessionState;
    function GetStateStr(): string;
    function GetFlags(): CK_ULONG;
    function GetDeviceError(): CK_ULONG;

    function GetReadOnly(): boolean;

    function Login(userType: TPKCS11UserType; usePINPad: boolean; PIN: string): boolean; overload;
    function GetLoggedIn(): boolean;

    function InitPIN(usePINPad: boolean; PIN: string): boolean; overload;
    function SetPIN(usePINPad: boolean; oldPIN: string; newPIN: string): boolean; overload;

    // Conversion functions...
    function CK_USER_TYPEToUserType(ckUserType: CK_USER_TYPE): TPKCS11UserType;
    function UserTypeToCK_USER_TYPE(userType: TPKCS11UserType): CK_USER_TYPE;
    function CK_STATEToState(ckState: CK_STATE): TPKCS11SessionState;
    function StateToCK_STATE(state: TPKCS11SessionState): CK_STATE;

  public
    FhSession: CK_SESSION_HANDLE;

    property Open: boolean read GetOpen;

    // Some PKCS#11 providers return incorrect slot ID in the info
    // This method returns the *real* slot ID...
    property SlotID: integer read FSlotID write FSlotID;
    // ...and this method returns the slot ID from the session's info from the
    // PKCS#11 API
    function SlotIDFromInfo(): integer;

    property State: TPKCS11SessionState read GetState;
    property StateStr: string read GetStateStr;
    property Flags: CK_FLAGS read GetFlags;
    property DeviceError: CK_ULONG read GetDeviceError;

    function GetSessionInfo(var info: CK_SESSION_INFO): boolean;

    property ReadOnly: boolean read GetReadOnly;

    function SeedRandom(byteCount: integer; seedData: string): boolean;
    function GenerateRandom(byteCount: integer; var randomData: string): boolean;

    function  Login(userType: TPKCS11UserType): boolean; overload;
    function  Login(userType: TPKCS11UserType; PIN: string): boolean; overload;
    procedure Logout();

    property  LoggedIn: boolean read GetLoggedIn;

    function  InitPIN(): boolean; overload;
    function  InitPIN(PIN: string): boolean; overload;
    function  SetPIN(): boolean; overload;
    function  SetPIN(oldPIN: string; newPIN: string): boolean; overload;

    function FindObjectsInit(): boolean;
    function FindNextObject(): TPKCS11Object;
    function FindObjectsFinal(): boolean;

    function GenerateKey(mechanism: CK_MECHANISM_TYPE; attrTemplate: TPKCS11AttributeTemplate): TPKCS11Object; overload;
    function CreateObject(attrObjs: array of TPKCS11Attribute): TPKCS11Object;
    function CopyObject(obj: TPKCS11Object): TPKCS11Object;
    function DestroyObject(obj: TPKCS11Object): boolean;

    function GetObject(application: string; objLabel: string): TPKCS11Object;
    function GetObjects(objClass: CK_OBJECT_CLASS): TPKCS11ObjectArray; overload;
    function GetObjects(application: string): TPKCS11ObjectArray; overload;
    function GetObjects(application: string; objLabel: string): TPKCS11ObjectArray; overload;

    function  EncryptInit(mechanism: CK_MECHANISM_TYPE; keyObj: TPKCS11Object): boolean; overload;
    function  EncryptInit(mechanism: TPKCS11Mechanism; keyObj: TPKCS11Object): boolean; overload;
    function  Encrypt(plaintext: string; var cyphertext: string): boolean;
    function  EncryptUpdate(plaintext: string; var cyphertext: string): boolean;
    function  EncryptFinal(var cyphertext: string): boolean;
    function  DecryptInit(mechanism: CK_MECHANISM_TYPE; keyObj: TPKCS11Object): boolean; overload;
    function  DecryptInit(mechanism: TPKCS11Mechanism; keyObj: TPKCS11Object): boolean; overload;
    function  Decrypt(cyphertext: string; var plaintext: string): boolean;
    function  DecryptUpdate(cyphertext: string; var plaintext: string): boolean;
    function  DecryptFinal(var plaintext: string): boolean;

    procedure CloseSession();
  end;


implementation

uses
  SysUtils,
  Shredder;

const
  PKCS11_IGNORE_PARAMETER = '!!IGNORE_THIS_PARAMETER!!';

function TPKCS11Session.GetSessionInfo(var info: CK_SESSION_INFO): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_GetSessionInfo, 'C_GetSessionInfo');

  LastRV := LibraryFunctionList.CK_C_GetSessionInfo(FhSession, @info);
  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.SeedRandom(byteCount: integer; seedData: string): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_SeedRandom, 'C_SeedRandom');

  LastRV := LibraryFunctionList.CK_C_SeedRandom(
                                               FhSession,
                                               CK_BYTE_PTR(PByte(seedData)),
                                               byteCount
                                              );

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.GenerateRandom(byteCount: integer; var randomData: string): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_GenerateRandom, 'C_GenerateRandom');

  // Setup the returned string so it's large enough to store the random data
  randomData := StringOfChar('X', byteCount);

  LastRV := LibraryFunctionList.CK_C_GenerateRandom(
                                                FhSession,
                                                CK_BYTE_PTR(PByte(randomData)),
                                                byteCount
                                               );

  Result := RVSuccess(LastRV);
end;

// Log user in using PIN pad
function TPKCS11Session.Login(userType: TPKCS11UserType): boolean;
begin
  Result := Login(userType, TRUE, '');
end;

// Log user in using specified PIN
function TPKCS11Session.Login(userType: TPKCS11UserType; PIN: string): boolean;
begin
  Result := Login(userType, FALSE, PIN);
end;

function TPKCS11Session.Login(userType: TPKCS11UserType; usePINPad: boolean; PIN: string): boolean;
var
  pkcs11UserType: CK_USER_TYPE;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_Login, 'C_Login');

  pkcs11UserType := UserTypeToCK_USER_TYPE(userType);

  if usePINPad then
    begin
    LastRV := LibraryFunctionList.CK_C_Login(FhSession, pkcs11UserType, NULL_PTR, 0);
    end
  else
    begin
    LastRV := LibraryFunctionList.CK_C_Login(
                                            FhSession,
                                            pkcs11UserType,
                                            CK_UTF8CHAR_PTR(PChar(PIN)),
                                            length(PIN)
                                           );
    end;
    
  Result := RVSuccess(LastRV);
end;

procedure TPKCS11Session.Logout();
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_Logout, 'C_Logout');
  LastRV := LibraryFunctionList.CK_C_Logout(FhSession);
end;

function TPKCS11Session.FindObjectsInit(): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_FindObjectsInit, 'C_FindObjectsInit');
  LastRV := LibraryFunctionList.CK_C_FindObjectsInit(FhSession, NULL_PTR, 0);
  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.GetLoggedIn(): boolean;
var
  currState: TPKCS11SessionState;
begin
  currState := State;
  Result := (
             Open and
             (currState <> ssROPublic) and
             (currState <> ssRWPublic)
            );
end;

function TPKCS11Session.GetReadOnly(): boolean;
var
  currState: TPKCS11SessionState;
begin
  currState := State;
  Result := (
             not(Open) or
             (currState = ssROPublic) or
             (currState = ssROUserFunctions)
            );
end;

// Important: It is the CALLERS responsibility to free off the object returned
// Returns nil if no more/error
function TPKCS11Session.FindNextObject(): TPKCS11Object;
var
  hObject: CK_OBJECT_HANDLE;
  ulObjectCount: CK_ULONG;
  retval: TPKCS11Object;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_FindObjects, 'C_FindObjects');

  retval := nil;
  
  LastRV := LibraryFunctionList.CK_C_FindObjects(
                                                FhSession,
                                                @hObject,
                                                1,
                                                @ulObjectCount
                                               );
  if (
      RVSuccess(LastRV) and
      (ulObjectCount <> 0)
     ) then
    begin
    retval := TPKCS11Object.Create();
    retval.LibraryFunctionList := LibraryFunctionList;
    retval.FSessionHandle := FhSession;
    retval.FHandle := hObject;
    end;
    
  Result := retval;
end;

function TPKCS11Session.FindObjectsFinal(): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_FindObjectsFinal, 'C_FindObjectsFinal');
  LastRV := LibraryFunctionList.CK_C_FindObjectsFinal(FhSession);
  Result := RVSuccess(LastRV);
end;


function TPKCS11Session.CreateObject(attrObjs: array of TPKCS11Attribute): TPKCS11Object;
var
  retval: TPKCS11Object;
  newObjHandle: CK_OBJECT_HANDLE;
  ckAttributes: array of CK_ATTRIBUTE;
  bufferArray: array of string;
  i: integer;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_CreateObject, 'C_CreateObject');

  retval := nil;

  setlength(ckAttributes, length(attrObjs));
  setlength(bufferArray, length(attrObjs));
  for i:=low(attrObjs) to high(attrObjs) do
    begin
    bufferArray[i] := attrObjs[i].ValueAsRawData;

    ckAttributes[i].attrType := attrObjs[i].AttribType;
    ckAttributes[i].pValue := PChar(bufferArray[i]);
    ckAttributes[i].ulValueLen := length(bufferArray[i]);
    end;

  LastRV := LibraryFunctionList.CK_C_CreateObject(
                                               FhSession,
                                               @(ckAttributes[0]),
                                               length(attrObjs),
                                               @newObjHandle
                                              );
  if RVSuccess(LastRV) then
    begin
    retval := TPKCS11Object.Create();
    retval.LibraryFunctionList := LibraryFunctionList;
    retval.FSessionHandle := FhSession;
    retval.FHandle := newObjHandle;
    end;

  Result := retval;
end;

function TPKCS11Session.GenerateKey(mechanism: CK_MECHANISM_TYPE; attrTemplate: TPKCS11AttributeTemplate): TPKCS11Object;
var
  retval: TPKCS11Object;
  newObjHandle: CK_OBJECT_HANDLE;
  ckAttributes: array of CK_ATTRIBUTE;
  bufferArray: array of string;
  i: integer;
  mechStruct: CK_MECHANISM;
  attrPtr: CK_ATTRIBUTE_PTR;
  currTemplateAttr: TPKCS11Attribute;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_GenerateKey, 'C_GenerateKey');

  retval := nil;

  mechStruct.mechanism      := mechanism;
  mechStruct.pParameter     := NULL_PTR;
  mechStruct.ulParameterLen := 0;

  attrPtr := NULL_PTR;
  if (attrTemplate.Count > 0) then
    begin
    setlength(ckAttributes, attrTemplate.Count);
    setlength(bufferArray, attrTemplate.Count);
    for i:=0 to (attrTemplate.Count - 1) do
      begin
      currTemplateAttr:= attrTemplate.Attribute[i];

      bufferArray[i] := currTemplateAttr.ValueAsRawData;

      ckAttributes[i].attrType := currTemplateAttr.AttribType;
      ckAttributes[i].pValue := PChar(bufferArray[i]);
      ckAttributes[i].ulValueLen := length(bufferArray[i]);
      end;

    attrPtr := @(ckAttributes[0]);
    end;

  LastRV := LibraryFunctionList.CK_C_GenerateKey(
                                               FhSession,
                                               @mechStruct,
                                               attrPtr,
                                               attrTemplate.Count,
                                               @newObjHandle
                                              );
  if RVSuccess(LastRV) then
    begin
    retval := TPKCS11Object.Create();
    retval.LibraryFunctionList := LibraryFunctionList;
    retval.FSessionHandle := FhSession;
    retval.FHandle := newObjHandle;
    end;

  Result := retval;
end;

function TPKCS11Session.CopyObject(obj: TPKCS11Object): TPKCS11Object;
var
  retval: TPKCS11Object;
  newObjHandle: CK_OBJECT_HANDLE;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_CopyObject, 'C_CopyObject');

  retval := nil;
  LastRV := LibraryFunctionList.CK_C_CopyObject(
                                               FhSession,
                                               obj.FHandle,
                                               nil,
                                               0,
                                               @newObjHandle
                                              );
  if RVSuccess(LastRV) then
    begin
    retval := TPKCS11Object.Create();
    retval.LibraryFunctionList := LibraryFunctionList;
    retval.FSessionHandle := FhSession;
    retval.FHandle := newObjHandle;
    end;

  Result := retval;
end;

function TPKCS11Session.DestroyObject(obj: TPKCS11Object): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_DestroyObject, 'C_DestroyObject');
  LastRV := LibraryFunctionList.CK_C_DestroyObject(FhSession, obj.FHandle);
  Result := RVSuccess(LastRV);
end;

// Returns -1 on error
function TPKCS11Session.SlotIDFromInfo(): integer;
var
  retval: integer;
  info: CK_SESSION_INFO;
begin
  retval := -1;

  if GetSessionInfo(info) then
    begin
    retval := info.SlotID;
    end;

  Result := retval;
end;

function TPKCS11Session.GetState(): TPKCS11SessionState;
var
  retval: TPKCS11SessionState;
  info: CK_SESSION_INFO;
begin
  retval := ssROPublic;

  if GetSessionInfo(info) then
    begin
    retval := CK_STATEToState(info.State);
    end;

  Result := retval;
end;

function TPKCS11Session.GetStateStr(): string;
begin
  Result := PKCS11_SESSION_STATE_DESC[State];
end;

function TPKCS11Session.GetFlags(): CK_ULONG;
var
  retval: CK_FLAGS;
  info: CK_SESSION_INFO;
begin
  retval := 0;

  if GetSessionInfo(info) then
    begin
    retval := info.Flags;
    end;

  Result := retval;
end;

function TPKCS11Session.GetDeviceError(): CK_ULONG;
var
  retval: CK_ULONG;
  info: CK_SESSION_INFO;
begin
  retval := 0;

  if GetSessionInfo(info) then
    begin
    retval := info.ulDeviceError;
    end;

  Result := retval;
end;


function TPKCS11Session.UserTypeToCK_USER_TYPE(userType: TPKCS11UserType): CK_USER_TYPE;
begin
  Result := PKCS11_USER_TYPE[userType];
end;

function TPKCS11Session.CK_USER_TYPEToUserType(ckUserType: CK_USER_TYPE): TPKCS11UserType;
var
  retval: TPKCS11UserType;
  i: TPKCS11UserType;
begin
  retval := utUser;

  for i:=low(PKCS11_USER_TYPE) to high(PKCS11_USER_TYPE) do
    begin
    if (PKCS11_USER_TYPE[i] = ckUserType) then
      begin
      retval := i;
      break
      end;
    end;

  Result := retval;
end;

function TPKCS11Session.StateToCK_STATE(state: TPKCS11SessionState): CK_STATE;
begin
  Result := PKCS11_SESSION_STATE[state];
end;

function TPKCS11Session.CK_STATEToState(ckState: CK_STATE): TPKCS11SessionState;
var
  retval: TPKCS11SessionState;
  i: TPKCS11SessionState;
begin
  retval := ssROPublic;

  for i:=low(PKCS11_SESSION_STATE) to high(PKCS11_SESSION_STATE) do
    begin
    if (PKCS11_SESSION_STATE[i] = ckState) then
      begin
      retval := i;
      break
      end;
    end;

  Result := retval;
end;

// Returns TRUE/FALSE, depending on whether the session is open/valid or
// not
function TPKCS11Session.GetOpen(): boolean;
var
  junk: CK_SESSION_INFO;
begin
  Result := GetSessionInfo(junk);
end;

function TPKCS11Session.InitPIN(usePINPad: boolean; PIN: string): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_InitPIN, 'C_InitPIN');

  LastRV := LibraryFunctionList.CK_C_InitPIN(
                                            FhSession,
                                            CK_UTF8CHAR_PTR(PChar(PIN)),
                                            length(PIN)
                                           );

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.SetPIN(usePINPad: boolean; oldPIN: string; newPIN: string): boolean;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_SetPIN, 'C_SetPIN');

  LastRV := LibraryFunctionList.CK_C_SetPIN(
                                           FhSession,
                                           CK_UTF8CHAR_PTR(PChar(oldPIN)),
                                           length(oldPIN),
                                           CK_UTF8CHAR_PTR(PChar(newPIN)),
                                           length(newPIN)
                                          );

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.InitPIN(): boolean;
begin
  Result := InitPIN(TRUE, '');
end;

function TPKCS11Session.InitPIN(PIN: string): boolean;
begin
  Result := InitPIN(FALSE, PIN);
end;

function TPKCS11Session.SetPIN(): boolean;
begin
  Result := SetPIN(TRUE, '', '');
end;

function TPKCS11Session.SetPIN(oldPIN: string; newPIN: string): boolean;
begin
  Result := SetPIN(FALSE, oldPIN, newPIN);
end;

// Retrieve object with the application/label attributes set to those
// specified
// Note: *Caller* is responsible for freeing returned attribute object
function TPKCS11Session.GetObject(application: string; objLabel: string): TPKCS11Object;
var
  tmpObjs: TPKCS11ObjectArray;
  retval: TPKCS11Object;
  i: integer;
begin
  retval := nil;

  tmpObjs := GetObjects(application, objLabel);
  if (length(tmpObjs) > 0) then
    begin
    retval := tmpObjs[low(tmpObjs)];

    // Ditch all remaining objects
    for i:=(low(tmpObjs) + 1) to high(tmpObjs) do
      begin
      tmpObjs[i].Free();
      end;
    end;

  Result := retval;
end;

// Retrieve object with the application/label attributes set to those
// specified
// Note: *Caller* is responsible for freeing *ALL* returned attribute objects
function TPKCS11Session.GetObjects(application: string): TPKCS11ObjectArray;
begin
  Result := GetObjects(application, PKCS11_IGNORE_PARAMETER);
end;

// Retrieve object with the application/label attributes set to those
// specified
// Note: *Caller* is responsible for freeing *ALL* returned attribute objects
function TPKCS11Session.GetObjects(application: string; objLabel: string): TPKCS11ObjectArray;
var
  retval: TPKCS11ObjectArray;
  currObj: TPKCS11Object;
  attrApp: TPKCS11Attribute;
  attrLabel: TPKCS11Attribute;
begin
  SetLength(retval, 0);
  
  if FindObjectsInit() then
    begin
    currObj := FindNextObject();
    while (currObj <> nil) do
      begin
      attrApp:= currObj.GetAttribute(CKA_APPLICATION);
      attrLabel:= currObj.GetAttribute(CKA_LABEL);

      if (
          (attrApp <> nil) and
          (attrLabel <> nil)
         ) then
        begin
        if (
            (
             (application = PKCS11_IGNORE_PARAMETER) or
             (trim(attrApp.ValueAsString) = trim(application))
            ) and
            (
             (objLabel = PKCS11_IGNORE_PARAMETER) or
             (trim(attrLabel.ValueAsString) = trim(objLabel))
            )
           ) then
          begin
          SetLength(retval, (length(retval) + 1));
          retval[length(retval)-1] := currObj;
          end
        else
          begin
          currObj.Free();
          end;
        end;

      if (attrApp <> nil) then
        begin
        attrApp.Free();
        end;
      if (attrLabel <> nil) then
        begin
        attrLabel.Free();
        end;

      currObj := FindNextObject();
      end;

    FindObjectsFinal();
    end;

  Result := retval;
end;

function TPKCS11Session.GetObjects(objClass: CK_OBJECT_CLASS): TPKCS11ObjectArray;
var
  retval: TPKCS11ObjectArray;
  currObj: TPKCS11Object;
  attrObjClass: TPKCS11Attribute;
begin
  SetLength(retval, 0);

  if FindObjectsInit() then
    begin
    currObj := FindNextObject();
    while (currObj <> nil) do
      begin
      attrObjClass:= currObj.GetAttribute(CKA_CLASS);

      if (attrObjClass <> nil) then
        begin
        if (attrObjClass.ValueAsNumber = objClass) then
          begin
          SetLength(retval, (length(retval) + 1));
          retval[length(retval)-1] := currObj;
          end
        else
          begin
          currObj.Free();
          end;
        end;

      if (attrObjClass <> nil) then
        begin
        attrObjClass.Free();
        end;

      currObj := FindNextObject();
      end;

    FindObjectsFinal();
    end;

  Result := retval;
end;

// Important: Keep implementation of:
//
//              TPKCS11Token.CloseSession(...);
//              TPKCS11Session.CloseSession();
//
//            in sync
procedure TPKCS11Session.CloseSession();
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_CloseSession, 'C_CloseSession');
  LastRV := LibraryFunctionList.CK_C_CloseSession(FhSession);
end;

function TPKCS11Session.EncryptInit(mechanism: CK_MECHANISM_TYPE; keyObj: TPKCS11Object): boolean;
var
  mechStruct: CK_MECHANISM;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_EncryptInit, 'C_EncryptInit');

  mechStruct.mechanism      := mechanism;
  mechStruct.pParameter     := NULL_PTR;
  mechStruct.ulParameterLen := 0;

  LastRV := LibraryFunctionList.CK_C_EncryptInit(
                                                 FhSession,
                                                 @mechStruct,
                                                 keyObj.FHandle
                                                );

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.EncryptInit(mechanism: TPKCS11Mechanism; keyObj: TPKCS11Object): boolean;
begin
  Result := EncryptInit(mechanism.MechanismType, keyObj);
end;

function TPKCS11Session.Encrypt(plaintext: string; var cyphertext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpCyphertext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_Encrypt, 'C_Encrypt');

  tmpLen := length(plaintext);
  FLastEncryptDecryptLen := tmpLen;
  tmpCyphertext := StringOfChar(#0, tmpLen);

  LastRV := LibraryFunctionList.CK_C_Encrypt(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(plaintext)),
                                             length(plaintext),
                                             CK_BYTE_PTR(PChar(tmpCyphertext)),
                                             @tmpLen
                                            );

  cyphertext := Copy(tmpCyphertext, 1, tmpLen);
  Overwrite(tmpCyphertext);

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.EncryptUpdate(plaintext: string; var cyphertext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpCyphertext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_EncryptUpdate, 'C_EncryptUpdate');

  tmpLen := length(plaintext);
  FLastEncryptDecryptLen := tmpLen;
  tmpCyphertext := StringOfChar(#0, tmpLen);

  LastRV := LibraryFunctionList.CK_C_EncryptUpdate(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(plaintext)),
                                             length(plaintext),
                                             CK_BYTE_PTR(PChar(tmpCyphertext)),
                                             @tmpLen
                                            );

  cyphertext := Copy(tmpCyphertext, 1, tmpLen);
  Overwrite(tmpCyphertext);

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.EncryptFinal(var cyphertext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpCyphertext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_EncryptFinal, 'C_EncryptFinal');

  tmpCyphertext := StringOfChar(#0, FLastEncryptDecryptLen);

  LastRV := LibraryFunctionList.CK_C_EncryptFinal(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(tmpCyphertext)),
                                             @tmpLen
                                            );

  cyphertext := Copy(tmpCyphertext, 1, tmpLen);
  Overwrite(tmpCyphertext);

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.DecryptInit(mechanism: CK_MECHANISM_TYPE; keyObj: TPKCS11Object): boolean;
var
  mechStruct: CK_MECHANISM;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_DecryptInit, 'C_DecryptInit');

  mechStruct.mechanism      := mechanism;
  mechStruct.pParameter     := NULL_PTR;
  mechStruct.ulParameterLen := 0;

  LastRV := LibraryFunctionList.CK_C_DecryptInit(
                                                 FhSession,
                                                 @mechStruct,
                                                 keyObj.FHandle
                                                );

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.DecryptInit(mechanism: TPKCS11Mechanism; keyObj: TPKCS11Object): boolean;
begin
  Result := DecryptInit(mechanism.MechanismType, keyObj);
end;

function TPKCS11Session.Decrypt(cyphertext: string; var plaintext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpPlaintext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_Decrypt, 'C_Decrypt');

  tmpLen := length(cyphertext);
  FLastEncryptDecryptLen := tmpLen;
  tmpPlaintext := StringOfChar(#0, tmpLen);

  LastRV := LibraryFunctionList.CK_C_Decrypt(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(cyphertext)),
                                             length(cyphertext),
                                             CK_BYTE_PTR(PChar(tmpPlaintext)),
                                             @tmpLen
                                            );

  plaintext := Copy(tmpPlaintext, 1, tmpLen);
  Overwrite(tmpPlaintext);

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.DecryptUpdate(cyphertext: string; var plaintext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpPlaintext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_DecryptUpdate, 'C_DecryptUpdate');

  tmpLen := length(cyphertext);
  FLastEncryptDecryptLen := tmpLen;
  tmpPlaintext := StringOfChar(#0, tmpLen);

  LastRV := LibraryFunctionList.CK_C_DecryptUpdate(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(cyphertext)),
                                             length(cyphertext),
                                             CK_BYTE_PTR(PChar(tmpPlaintext)),
                                             @tmpLen
                                            );

  plaintext := Copy(tmpPlaintext, 1, tmpLen);
  Overwrite(tmpPlaintext);

  Result := RVSuccess(LastRV);
end;

function TPKCS11Session.DecryptFinal(var plaintext: string): boolean;
var
  tmpLen: CK_ULONG;
  tmpPlaintext: string;
begin
  CheckFnAvailable(@LibraryFunctionList.CK_C_DecryptFinal, 'C_DecryptFinal');

  tmpPlaintext := StringOfChar(#0, FLastEncryptDecryptLen);

  LastRV := LibraryFunctionList.CK_C_DecryptFinal(
                                             FhSession,
                                             CK_BYTE_PTR(PChar(tmpPlaintext)),
                                             @tmpLen
                                            );

  plaintext := Copy(tmpPlaintext, 1, tmpLen);
  Overwrite(tmpPlaintext);

  Result := RVSuccess(LastRV);
end;

END.

